Require Import List ListSet Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.AbstractValues
        CSL.Analysis.SingletonSet.

Set Implicit Arguments.

Definition abstract_agent := abstract_value Agent.
Definition defparticip_result := set abstract_agent.

Definition dec_eq_aa : forall (x y : abstract_agent), {x = y} + {x <> y}.
  repeat decide equality.
Defined.

Program Definition ev_to_aa e : abstract_agent :=
  match e with
  | Event a₁ _ _ _ => ActualValue a₁
  end.

Definition β_defparticip_folder (e : event) (s: defparticip_result) : defparticip_result :=
  set_add dec_eq_aa (ev_to_aa e) s.

Definition β_defparticip (t: trace) : defparticip_result :=
  fold_right β_defparticip_folder (empty_set _) t.

Hint Resolve set_union_elim set_union_intro set_union_intro1 set_union_intro2 set_add_elim set_add_intro set_add_intro1 set_add_intro2 : csl.
Hint Unfold β_defparticip_folder : csl.

Lemma β_defparticip_app_inter : forall (t1 t2 t3 : trace) a,
    appends t1 t2 t3 ->
    set_In a (set_inter dec_eq_aa (β_defparticip t1) (β_defparticip t2)) -> set_In a (β_defparticip t3).
Proof.
  intros; induction H; simpl in *; autounfold in *; intuition.
  apply set_add_intro.
  apply set_inter_elim in H0. destruct H0.
  apply set_add_elim in H0; destruct H0; intuition.
Qed.

Lemma β_defparticip_app_union : forall (t1 t2 t3 : trace) a,
    appends t1 t2 t3 ->
    set_In a (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)) -> set_In a (β_defparticip t3).
Proof.
  intros; induction H; simpl in *; autounfold with csl in *; intuition;
    apply set_union_elim in H0; intuition.
  apply set_add_elim in H1; intuition.
Qed.

Lemma β_defparticip_interleave_inter : forall (t1 t2 t3 : trace) a,
    interleave t1 t2 t3 ->
    set_In a (set_inter dec_eq_aa (β_defparticip t1) (β_defparticip t2)) -> set_In a (β_defparticip t3).
Proof.
  intros; induction H; simpl in *; autounfold with csl in *; intuition;
  apply set_inter_elim in H0; destruct H0; intuition.
  apply set_add_elim in H0; destruct H0; intuition.
  apply set_add_elim in H1; destruct H1; intuition.
Qed.

Lemma β_defparticip_interleave_union : forall (t1 t2 t3 : trace) a,
    interleave t1 t2 t3 ->
    set_In a (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)) -> set_In a (β_defparticip t3).
Proof.
  intros; induction H; simpl in *; autounfold with csl in *; intuition;
  apply set_union_elim in H0; destruct H0; intuition;
  apply set_add_elim in H0; destruct H0; intuition.
Qed.

Definition aa_set_In (aa : abstract_agent) (s : set abstract_agent) : Prop :=
  match aa with
  | AnyValue => set_In aa s
  | ActualValue v => set_In aa s \/ set_In AnyValue s
  end.

Lemma aa_set_β_defparticip:
  forall a s,
    aa_set_In a (β_defparticip s) -> set_In a (β_defparticip s).
Proof.
  induction s; intros; destruct a; simpl in *; intuition;
  unfold β_defparticip_folder in *;
  destruct a0; simpl in *; apply set_add_elim in H0; destruct H0;
  intuition; discriminate.
Qed.

Lemma β_defparticip_only_concrete_agents:
  forall t v,
    aa_set_In v (β_defparticip t) -> exists a, v = ActualValue a.
Proof.
  induction t; unfold aa_set_In; simpl; destruct v; simpl in *; intuition; unfold β_defparticip_folder in *.
  + apply set_add_elim in H; destruct a; simpl in *; intuition; discriminate.
  + exists t0; auto.
  + exists t0; auto.
Qed.

Definition aa_set_InclRev (s1 s2 : set abstract_agent) :=
  forall a, aa_set_In a s2 -> aa_set_In a s1.

Fixpoint aa_set_mem (a : abstract_agent) (s: set abstract_agent) : bool :=
  match s with
  | nil => false
  | a1 :: s' =>
    match a1 with
    | AnyValue => true
    | ActualValue t =>
      match dec_eq_aa a a1 with
      | left _ => true
      | right _ => aa_set_mem a s'
      end
    end
  end.

Fixpoint aa_set_inter (s1 : set abstract_agent) : set abstract_agent -> set abstract_agent :=
  match s1 with
  | nil => fun s2 => nil
  | a1 :: s1' =>
    match a1 with
    | AnyValue => fun s2 => s2
    | ActualValue t =>
      fun s2 =>
        if aa_set_mem a1 s2 then a1 :: aa_set_inter s1' s2 else aa_set_inter s1' s2
    end
  end.

Definition singleton_abstract (s: abstract_set Agent) : set abstract_agent :=
  match s with
  | FullSet => singleton dec_eq_aa AnyValue
  | ActualSet s => set_map dec_eq_aa (fun v => ActualValue v) s
  end.

Lemma aa_set_mem_correct:
  forall a s, set_In a s -> aa_set_mem a s = true.
Proof.
  induction s; simpl in *; intuition; destruct (dec_eq_aa a a0); subst; try destruct a; try destruct a0; intuition.
Qed.

Lemma aa_set_mem_aa_set_In:
  forall a s, aa_set_In a s <-> aa_set_mem a s = true.
Proof.
  induction s; simpl in *; destruct a; intuition; simpl in H; intuition.
  + destruct a0; simpl in *; intuition; discriminate.
  + destruct a0; simpl in *; intuition.
  + destruct a0; intuition; destruct (dec_eq_aa (ActualValue t) (ActualValue t0)); intuition; simpl in H1; intuition; discriminate.
  + destruct a0; simpl; intuition; destruct (dec_eq_aa (ActualValue t) (ActualValue t0)); intuition; simpl in H; intuition.
Qed.

Lemma aa_set_mem_AnyValue_spec:
  forall s, set_In AnyValue s -> forall t, aa_set_mem (ActualValue t) s = true.
Proof.
  intros; apply aa_set_mem_aa_set_In; simpl; intuition.
Qed.

Lemma aa_set_union_intro : forall a s1 s2,
    aa_set_In a s1 \/ aa_set_In a s2 -> aa_set_In a (set_union dec_eq_aa s1 s2).
Proof.
  intros; unfold aa_set_In in *; destruct a; intuition.
Qed.

Lemma aa_set_union_elim : forall a s1 s2,
    aa_set_In a (set_union dec_eq_aa s1 s2) -> aa_set_In a s1 \/ aa_set_In a s2.
Proof.
  intros; unfold aa_set_In in *; destruct a;
  try apply set_union_elim in H; intuition; apply set_union_elim in H0; intuition.
Qed.

Lemma aa_set_inter_intro : forall s1 s2 a,
    aa_set_In a s1 /\ aa_set_In a s2 -> aa_set_In a (aa_set_inter s1 s2).
Proof.
  intros; destruct a; intuition.
  - simpl; induction s1; simpl in *; intuition; subst; try destruct a; auto;
    destruct (aa_set_mem (@ActualValue Agent t) s2); simpl; intuition.
  - simpl in *; intuition.
    + left; induction s1; simpl in *; intuition; subst.
      * apply aa_set_mem_correct in H0; rewrite H0; simpl; intuition.
      * destruct a; auto; destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl; intuition.
    + assert (aa_set_mem (@ActualValue Agent t) s2 = true)
        by (apply aa_set_mem_AnyValue_spec; assumption);
      induction s1; simpl in *; intuition; subst.
      * rewrite H1; simpl; intuition.
      * destruct a; auto;
          destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl; intuition.
      * destruct a; intuition; destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl; intuition.
    + assert (aa_set_mem (@ActualValue Agent t) s1 = true)
        by (apply aa_set_mem_AnyValue_spec; assumption);
      induction s1; simpl; simpl in H, H0; intuition; subst; intuition.
      destruct a; auto; destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl; intuition.
      * apply aa_set_mem_aa_set_In in H1; simpl in H1; intuition;
        apply aa_set_mem_AnyValue_spec with (t:=t) in H2; intuition.
      * apply aa_set_mem_aa_set_In in H1; simpl in H1; intuition;
          apply aa_set_mem_AnyValue_spec with (t:=t) in H2; intuition.
    + right. induction s1; simpl in *; intuition; subst; try destruct a; auto;
      destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl; intuition.
Qed.

Lemma aa_set_inter_elim_1 : forall a s1 s2,
    aa_set_In a (aa_set_inter s1 s2) -> aa_set_In a s1.
Proof.
  induction s1; simpl; intros; destruct a; intuition.
  - destruct a0; simpl; intuition;
      destruct (aa_set_mem (ActualValue t) s2); simpl in *; right; intuition;
        try discriminate; eapply IHs1; eauto.
  - destruct a0; simpl; intuition; simpl in *; intuition;
      destruct (aa_set_mem (@ActualValue Agent t0) s2); simpl in *; intuition;
        destruct IHs1 with (s2:=s2); intuition.
Qed.

Lemma aa_set_inter_elim_2 : forall a s1 s2,
    aa_set_In a (aa_set_inter s1 s2) -> aa_set_In a s2.
Proof.
  induction s1; destruct a; simpl; [ tauto | tauto | | ]; intros;
    destruct a0; auto.
  - destruct (aa_set_mem (ActualValue t) s2); simpl in H; intuition; [ discriminate | | ]; eapply IHs1; assumption.
  - generalize H;
    generalize (aa_set_mem_aa_set_In (ActualValue t0) s2);
    destruct (aa_set_mem (ActualValue t0) s2); simpl in *;
    intros; intuition; try discriminate; inversion H1; subst; auto.
Qed.

Lemma aa_set_inter_elim : forall a s1 s2,
    aa_set_In a (aa_set_inter s1 s2) -> aa_set_In a s1 /\ aa_set_In a s2.
Proof.
  intros; split; [eapply aa_set_inter_elim_1 | eapply aa_set_inter_elim_2]; eauto.
Qed.

Lemma aa_set_InclRev_β_defparticip:
  forall t l,
  aa_set_InclRev (β_defparticip t) l ->
  ~(aa_set_In AnyValue l).
Proof.
  unfold aa_set_InclRev; simpl; intros.
  induction l; simpl; auto.
  intro. intuition.
  + subst. specialize H with AnyValue.
    simpl in H. assert (set_In AnyValue (β_defparticip t)). intuition.
    assert (exists a, AnyValue = @ActualValue Agent a).
    eapply β_defparticip_only_concrete_agents. eauto. destruct H1; discriminate.
  + apply IHl; intuition. apply H. destruct a0; simpl; intuition.
Qed.

Lemma aa_set_add_intro:
  forall a v s,
    a = v \/ aa_set_In a s -> aa_set_In a (set_add dec_eq_aa v s).
Proof.
  intros; destruct a; simpl in *; intuition.
Qed.

Lemma aa_set_add_elim0:
  forall (a v : abstract_agent) (s : set abstract_agent), aa_set_In a (set_add dec_eq_aa v s) -> a = v \/ aa_set_In a s \/ v = AnyValue.
Proof.
  intros; destruct a; simpl in *; intuition.
  + apply set_add_elim in H; intuition.
  + apply set_add_elim in H0; intuition.
  + apply set_add_elim in H0; intuition.
Qed.

Lemma aa_set_add_elim: forall  (a : abstract_agent) e (s : set abstract_agent),
    aa_set_In a (set_add dec_eq_aa (ev_to_aa e) s) -> a = ev_to_aa e \/ aa_set_In a s.
Proof.
  intros; unfold aa_set_In in *; destruct e; destruct a; simpl in *; intuition;
    try match goal with
          [ H : set_In _ (set_add _ _ _) |- _ ] => apply set_add_elim in H; intuition
        end; try discriminate.
Qed.
