Require Import List ListSet Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.Lattice
        CSL.Analysis.SingletonSet
        CSL.Analysis.AbstractValues
        CSL.Analysis.DefParticip
        CSL.Analysis.Generic
        CSL.Analysis.IdPredicateAnalysis
        CSL.Analysis.Algorithms.
Set Implicit Arguments.
Hint Unfold aa_set_InclRev: csl.

Program Instance aa_set_lattice : Lattice defparticip_result :=
  { top := empty_set abstract_agent;
    bot := singleton dec_eq_aa AnyValue;
    Incl := aa_set_InclRev;
    join := aa_set_inter
  }.
Next Obligation.
  unfold aa_set_InclRev; destruct x; intros; destruct a; try destruct a0; simpl in *; intuition.
Defined.
Next Obligation.
  unfold aa_set_InclRev; destruct x; intros; destruct a; try destruct a0; simpl in *; intuition.
Defined.
Next Obligation.
  auto with csl.
Defined.
Next Obligation.
  auto with csl.
Defined.
Next Obligation.
  unfold aa_set_InclRev; split; intros; destruct (aa_set_inter_elim a x y); auto with csl.
Defined.
Next Obligation.
  unfold aa_set_InclRev; intros; apply aa_set_inter_elim in H1; apply aa_set_inter_intro; intuition.
Defined.

Lemma β_defparticip_empty:
  aa_set_InclRev (β_defparticip []) (empty_set abstract_agent).
Proof.
  simpl; intros; auto with csl.
Qed.

Lemma β_append : forall t1 t2 t l1 l2,
    appends t1 t2 t -> aa_set_InclRev (β_defparticip t1) l1 /\ aa_set_InclRev (β_defparticip t2) l2 -> aa_set_InclRev (β_defparticip t) (set_union dec_eq_aa l1 l2).
Proof.
  unfold aa_set_InclRev; intuition. apply aa_set_union_elim in H0; intuition.
  - apply β_defparticip_app_union with (a:=a) in H.
    + apply aa_set_mem_aa_set_In; apply aa_set_mem_correct; assumption.
    + apply set_union_intro; apply H1 in H3; apply aa_set_β_defparticip in H3; intuition.
  - apply β_defparticip_app_union with (a:=a) in H.
    + apply aa_set_mem_aa_set_In; apply aa_set_mem_correct; assumption.
    + apply set_union_intro; apply H2 in H3; apply aa_set_β_defparticip in H3; intuition.
Qed.

Lemma β_append_Incl_R : forall t1 t2 t,
    appends t1 t2 t -> aa_set_InclRev (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)) (β_defparticip t) .
Proof.
  unfold aa_set_InclRev.
  intros.
  induction H; simpl in *.
  - simpl. apply aa_set_union_intro; intuition.
  - autounfold with csl in *.
    apply aa_set_union_intro. apply aa_set_add_elim in H0. intuition; subst.
    + left; apply aa_set_add_intro; intuition.
    + apply aa_set_union_elim in H0; intuition; left; apply aa_set_add_intro; auto.
Qed.

Lemma β_append_Incl_L : forall t1 t2 t,
    appends t1 t2 t -> aa_set_InclRev (β_defparticip t) (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)).
Proof.
  unfold aa_set_InclRev.
  intros.
  induction H; simpl in *.
  - apply aa_set_union_elim in H0; intuition;
    destruct a; simpl in *; intuition.
  - autounfold with csl in *.
    apply aa_set_union_elim in H0; intuition; [apply aa_set_add_elim in H1; intuition | ]; subst;
      apply aa_set_add_intro; auto; right; apply IHappends; apply aa_set_union_intro; auto.
Qed.

Lemma β_interleave : forall t1 t2 t l1 l2,
    interleave t1 t2 t -> aa_set_InclRev (β_defparticip t1) l1 /\ aa_set_InclRev (β_defparticip t2) l2 -> aa_set_InclRev (β_defparticip t) (set_union dec_eq_aa l1 l2).
Proof.
  unfold aa_set_InclRev; intuition. apply aa_set_union_elim in H0; intuition.
  - apply β_defparticip_interleave_union with (a:=a) in H.
    + apply aa_set_mem_aa_set_In; apply aa_set_mem_correct; assumption.
    + apply set_union_intro; apply H1 in H3; apply aa_set_β_defparticip in H3; intuition.
  - apply β_defparticip_interleave_union with (a:=a) in H.
    + apply aa_set_mem_aa_set_In; apply aa_set_mem_correct; assumption.
    + apply set_union_intro; apply H2 in H3; apply aa_set_β_defparticip in H3; intuition.
Qed.

Lemma β_interleave_Incl_R : forall t1 t2 t,
    interleave t1 t2 t -> aa_set_InclRev (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)) (β_defparticip t) .
Proof.
  unfold aa_set_InclRev.
  intros.
  induction H; simpl in *; auto.
  - simpl. apply aa_set_union_intro; intuition.
  - autounfold with csl in *.
    apply aa_set_union_intro. apply aa_set_add_elim in H0. intuition; subst.
    + left; apply aa_set_add_intro; intuition.
    + apply aa_set_union_elim in H0; intuition; left; apply aa_set_add_intro; auto.
  - autounfold with csl in *.
    apply aa_set_union_intro. apply aa_set_add_elim in H0. intuition; subst.
    + right; apply aa_set_add_intro; intuition.
    + apply aa_set_union_elim in H0; intuition; right; apply aa_set_add_intro; auto.
Qed.

Lemma β_interleave_Incl_L : forall t1 t2 t,
    interleave t1 t2 t -> aa_set_InclRev (β_defparticip t) (set_union dec_eq_aa (β_defparticip t1) (β_defparticip t2)).
Proof.
  unfold aa_set_InclRev.
  intros.
  induction H; simpl in *; auto.
  - apply aa_set_union_elim in H0; intuition;
    destruct a; simpl in *; intuition.
  - autounfold with csl in *.
    apply aa_set_union_elim in H0; intuition; [apply aa_set_add_elim in H1; intuition | ]; subst;
      apply aa_set_add_intro; auto; right; apply IHinterleave; apply aa_set_union_intro; auto.
  - autounfold with csl in *.
    apply aa_set_union_elim in H0; intuition; [ | apply aa_set_add_elim in H1; intuition]; subst;
      apply aa_set_add_intro; auto; right; apply IHinterleave; apply aa_set_union_intro; auto.
Qed.

Definition transfer_analysis Δ (m : value_map (new_var Δ)) l :=
  let agent1 := (hget m HFirst) in
  match agent1 with
  | FullSet => l
  | ActualSet s =>
    match s with
    | [] => singleton dec_eq_aa AnyValue (* Failing transfer *)
    | a::s' => if (is_singleton_set s) then set_add dec_eq_aa (ActualValue a) l else l
    end
  end.

Lemma transfer_correct `{@PredicateAnalysis _ _ abstract_set_setlattice} :
  forall Δ (δ : env Δ) (m' : value_map (new_var Δ)) l e t,
    aenv_correct (addEvent e δ) m' ->
    aa_set_InclRev (β_defparticip t) l ->
    aa_set_InclRev (β_defparticip (e :: t)) (transfer_analysis m' l).
Proof.
  intros.
  assert (~(aa_set_In AnyValue l)) by (eapply aa_set_InclRev_β_defparticip; eauto).
  unfold aa_set_InclRev in *. intros. destruct a.
  - (* AnyValue: failing transfer – this trace is not a satisfying one! *)
    remember (H0 Agent HFirst). destruct e. simpl in i.
    unfold transfer_analysis in H3. dependent destruction m'. simpl in i, H3.
    destruct a.
    + simpl in H2; contradiction.
    + destruct s.
      * simpl in *; contradiction.
      * destruct (is_singleton_set (t4 :: s)); simpl in H3.
        -- apply set_add_elim in H3; intuition; discriminate.
        -- simpl in *. contradiction.
  - (* ActualValue t0 *)
    unfold transfer_analysis in H3.
    dependent destruction m'; destruct e. simpl in H0. simpl in H3.
    destruct a; simpl in * |-.
    + simpl in *; intuition. apply aa_set_mem_correct in H4. apply aa_set_mem_aa_set_In in H4. apply H1 in H4.
      simpl in H4. unfold β_defparticip_folder. intuition.
    + destruct s.
      * remember (H0 Agent HFirst). simpl in i. contradiction.
      * destruct H3; simpl; unfold β_defparticip_folder; simpl.
        -- remember (H0 Agent HFirst). simpl in i.
           remember (is_singleton_set (s::s0)).
           destruct b.
           ++ apply set_add_elim in H3. left. apply set_add_intro. intuition.
              ** destruct i.
                 --- inversion H4; subst. tauto.
                 --- symmetry in Heqb. apply is_singleton_set_In_eq with (v:=t1) in Heqb; subst; simpl; intuition.
              ** right; apply aa_set_β_defparticip.
                 apply aa_set_mem_correct in H4. apply aa_set_mem_aa_set_In in H4. auto.
           ++ left; apply set_add_intro. right. apply aa_set_β_defparticip.
                 apply aa_set_mem_correct in H3. apply aa_set_mem_aa_set_In in H3. auto.
        -- destruct (is_singleton_set (s::s0)); [apply set_add_elim in H3; intuition; discriminate | contradiction].
Qed.

Lemma defparticip_union_monotone : forall l1 l2 l3 l4,
    aa_set_InclRev l1 l2 /\ aa_set_InclRev l3 l4 ->
    aa_set_InclRev (set_union dec_eq_aa l1 l3) (set_union dec_eq_aa l2 l4).
Proof.
  unfold aa_set_InclRev; simpl; intros; apply aa_set_union_intro; apply aa_set_union_elim in H0; intuition.
Qed.

Lemma transfer_analysis_monotone`{@PredicateAnalysis _ _ abstract_set_setlattice}:
  forall (Δ : list ty) (m m' : value_map (new_var Δ)) l l',
    aenv_Incl m m' -> aa_set_InclRev l l' -> aa_set_InclRev (transfer_analysis m l) (transfer_analysis m' l').
Proof.
  intros.
  unfold transfer_analysis.
  remember (H0 Agent HFirst).
  dependent destruction m'; dependent destruction m.
  simpl in *.
  destruct a; destruct a0.
  - assumption.
  - contradiction.
  - destruct s.
    + unfold aa_set_InclRev in *; intros; destruct a; simpl; intuition.
    + destruct (is_singleton_set (t :: s)).
      * unfold aa_set_InclRev in *; intros. apply aa_set_add_intro; intuition.
      * assumption.
  - destruct s; destruct s0.
    + unfold aa_set_InclRev in *; intros; assumption.
    + unfold aa_set_InclRev in *; intros. destruct a; simpl; intuition.
    + unfold aa_set_InclRev in *; destruct s; intros; simpl in *; contradiction (i t); intuition.
    + unfold aa_set_InclRev in *; intros; remember (is_singleton_set (t::s)); remember (is_singleton_set (t0::s0)); destruct b; destruct b0.
      * (* singleton, singleton *)
        assert (t = t0).
        -- simpl in i. destruct (i t). intuition. auto.
           symmetry in Heqb0. apply is_singleton_set_In_eq with (v:=t) in Heqb0; simpl; intuition.
        -- subst. apply aa_set_add_elim0 in H2. apply aa_set_add_intro. intuition; discriminate.
      * (* singleton, list *)
        simpl; apply aa_set_add_intro; intuition.
      * (* list, singleton *)
        symmetry in Heqb; apply not_singleton in Heqb. destruct Heqb; intuition.
        assert (set_In t (t::s)) by (simpl; auto).
        simpl in i. remember i.
        symmetry in Heqb0.
        destruct (i x); intuition.
        -- destruct (i t); intuition.
           ** subst. elim H5; reflexivity.
           ** subst. assert (set_In t (x :: s0)) by (simpl; auto).
              apply is_singleton_set_In_eq in H6; auto. contradiction.
        --  assert (set_In x (t0::s0)) by (simpl; auto).
            destruct (i t); intuition; subst.
            ** apply is_singleton_set_In_eq in H7; auto. symmetry in H7; contradiction.
            ** assert (set_In t (x :: s0)) by (simpl; auto).
               apply is_singleton_set_In_eq in H9; auto; [contradiction | ].
               apply is_singleton_set_In_eq in H7; subst; auto.
      * (* list, list *)
        auto.
Qed.

Instance defparticip_analysis `{P : @PredicateAnalysis _ _ abstract_set_setlattice } : CSLAnalysis aa_set_lattice P :=
  {
    β := β_defparticip;
    L__succ := top;
    β_empty := β_defparticip_empty;
    C__par := set_union dec_eq_aa;
    C__seq := set_union dec_eq_aa;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := defparticip_union_monotone;
    monotone_C__seq := defparticip_union_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.

Program Instance defparticip_id_predicate_analysis:
  CSLAnalysis aa_set_lattice id_predicate_analysis_abstract_set :=
  {
    β := β_defparticip;
    L__succ := top;
    β_empty := β_defparticip_empty;
    C__par := set_union dec_eq_aa;
    C__seq := set_union dec_eq_aa;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := defparticip_union_monotone;
    monotone_C__seq := defparticip_union_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.

Program Instance defparticip_possible_values :
  CSLAnalysis aa_set_lattice possible_values_predicate_analysis :=
  {
    β := β_defparticip;
    L__succ := top;
    β_empty := β_defparticip_empty;
    C__par := set_union dec_eq_aa;
    C__seq := set_union dec_eq_aa;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := defparticip_union_monotone;
    monotone_C__seq := defparticip_union_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.   

Require Import String.
Open Scope string_scope.

(* This is an autogenerated example created by a simple program that parses (the de-sugaring of)
   the CSL code given below, and produces the corresponding Coq AST, with the deadline given as
   an arbitrary integer (i.e. not number of seconds from 1/1/1970)

letrec
  sale[trusted, seller, buyer, goods, payment, deadline] =
    Transfer(buyer, trusted, payment, _).
    (Transfer(seller, buyer, goods, T | T < deadline).
    Transfer(trusted, seller, payment, T’ | True).Success
    + Transfer(trusted, buyer, payment, T | T > deadline).Success)
  in sale("3rd", "shop", "alice", 1 bike, 1000 EUR, 2019-09-01)
*)
Example exampleSell :=
 LetRec (Δ:=[]) (Γ := (cons ((cons Agent (cons Agent (cons Agent (cons Resource (cons Resource (cons Timestamp []))))))) []))
	(HCons (transfer (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext HFirst))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))) (alternate (transfer (Binop And (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Binop (Le TimestampCmp ) (Var (HNext (HNext (HNext HFirst)))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))))) (transfer (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))))))))) (@success _ _))) (transfer (Binop And (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))))) (Binop (Le TimestampCmp ) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Var (HNext (HNext (HNext HFirst)))))) (@success _ _)))) HNil)
	(tapp (HCons (@Lit _ Agent "3rd") (HCons (@Lit _ Agent "shop") (HCons (@Lit _ Agent "alice") (HCons (@Lit _ Resource [("bike",1)]) (HCons (@Lit _ Resource [("eur",1000)]) (HCons (@Lit _ Timestamp 2019) HNil)))))) HFirst).

Eval compute in (analyzeTopLevel (CA:=defparticip_possible_values) 500 exampleSell HNil).
