Require Import List ListSet Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.Lattice
        CSL.Analysis.AbstractValues
        CSL.Analysis.Interval
        CSL.Analysis.Fairness
        CSL.Analysis.Generic.

Set Implicit Arguments.
  
Lemma β_append_eq : forall t1 t2 t,
    appends t1 t2 t ->
    forall a : tyDenote Agent, eq_interval (plus_interval (β_fairness t1 a) (β_fairness t2 a)) (β_fairness t a).
Proof.
  intros.
  induction H. 
  - apply plus_interval_l_zero.
  - simpl. unfold β_fairness_folder. unfold add_fr.
    rewrite plus_interval_assoc.
    apply plus_interval_cong_r. assumption.
Qed.

Lemma β_interleave_eq : forall t1 t2 t,
    interleave t1 t2 t ->
    forall a : tyDenote Agent, eq_interval (plus_interval (β_fairness t1 a) (β_fairness t2 a)) (β_fairness t a).
Proof.
  intros.
  induction H. 
  - apply plus_interval_l_zero.
  - apply plus_interval_r_zero.
  - simpl. unfold β_fairness_folder. unfold add_fr. 
    rewrite plus_interval_assoc.
    apply plus_interval_cong_r. assumption.
  - simpl. unfold β_fairness_folder. unfold add_fr.
    rewrite plus_interval_comm.
    rewrite plus_interval_assoc.
    apply plus_interval_cong_r.
    rewrite plus_interval_comm.
    assumption.
Qed.

Program Instance fairness_analysis `{P : @PredicateAnalysis _ _ abstract_set_setlattice } :
  CSLAnalysis (@fun_lattice (tyDenote Agent) interval _) P :=
  {
    β := β_fairness;
    L__succ := success_fr;
    C__par := add_fr; C__seq := add_fr;
    C__transfer {Δ} m l := add_fr (transfer_analysis m) l
  }.
Next Obligation.
  autounfold with csl; auto.
Defined.
Next Obligation. (* β_par *)
  specialize (H0 a).
  eapply β_fairness_interleaves_add_eq in H; auto.
  unfold add_fr in *.
  eapply Incl_interval_trans.
  apply Incl_interval_eq.
  apply H.
  apply plus_interval_monotone; auto.
Defined.
Next Obligation. (* monotone_C_par *)
  intros.
  simpl in *.
  specialize (H a).
  intros.
  unfold add_fr in *.
  apply plus_interval_monotone; auto.
Defined.
Next Obligation. (* β_seq *)
  simpl; intros.
  specialize (H0 a).
  eapply β_fairness_app_add_eq in H; auto.
  unfold add_fr in *.
  eapply Incl_interval_trans.
  apply Incl_interval_eq.
  apply H.
  apply plus_interval_monotone; auto.
Defined.
Next Obligation. (* monotone_C_seq *)
  intros.
  simpl in *.
  specialize (H a).
  intros.
  unfold add_fr in *.
  apply plus_interval_monotone; auto.
Defined.
Next Obligation. (* β_transfer *)
  intros.
  simpl in *.
  intros.
  eapply transfer_analysis_correct with (a := a) in H.
  unfold β_fairness_folder.
  unfold add_fr in *.
  apply plus_interval_monotone; auto.
Defined.
Next Obligation. (* monotone_C_transfer *)
  intros.
  simpl in *.
  intros.
  unfold add_fr in *.
  apply plus_interval_monotone.

  unfold transfer_analysis.
  unfold add_fr.

  dependent destruction m.
  dependent destruction m.
  dependent destruction m.
  dependent destruction m'.
  dependent destruction m'.
  dependent destruction m'.
  simpl.

  remember (H Agent HFirst) as H1.
  remember (H Agent (HNext HFirst)) as H2.
  remember (H Resource (HNext (HNext HFirst))) as H3.

  simpl in *.
  clear HeqH3 HeqH2 HeqH1.
  apply plus_interval_monotone;
  apply value_transfer_monotone; auto.
  apply H0.
Defined.
