Require Import Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.AbstractValues
        CSL.Analysis.Lattice.

Set Implicit Arguments.

Class PredicateAnalysis (A : ty -> Type) `(L : SetLattice A) :=
  {
    arg_eval : forall Δ τ, exp Δ τ -> hlist A Δ -> A τ;

    aenv_correct {Δ} (δ : env Δ) (m : hlist A Δ) :=
      forall τ (x : member τ Δ), In (hget δ x) (hget m x);

    aenv_Incl {Δ} (m m' : hlist A Δ) :=
      forall τ (x : member τ Δ), Incl (hget m x) (hget m' x);

    arg_env_aenv := fix F {Δ ts} (m : hlist A Δ) (ae : arg_env ts Δ) : hlist A ts :=
      match ae with
      | HNil => HNil
      | HCons e ae' => HCons (arg_eval e m) (F m ae')
      end;

    add_event_aenv {Δ} (m : hlist A Δ) : hlist A (new_var Δ) :=
      HCons top (HCons top (HCons top (HCons top m)));

    arg_eval_correct : forall Δ (δ : env Δ) (m : hlist A Δ) τ (e : exp Δ τ),
        aenv_correct δ m -> In (expDenote e δ) (arg_eval e m);

    arg_eval_monotone : forall Δ (m m' : hlist A Δ) τ (e : exp Δ τ),
        aenv_Incl m m' -> Incl (arg_eval e m) (arg_eval e m');

    analysis : forall Δ, hlist A (new_var Δ) -> exp (new_var Δ) Bool -> option (hlist A (new_var Δ));

    analysis_dec : forall Δ (p : exp (new_var Δ) Bool) m,
        { m' | analysis m p = Some m' } + { analysis m p = None };

    analysis_None : forall Δ (p : exp (new_var Δ) Bool) m δ,
        analysis m p = None /\ aenv_correct δ m -> expDenote p δ = false;

    analysis_Some : forall Δ (p : exp (new_var Δ) Bool) m m' δ,
        aenv_correct δ m /\ analysis m p = Some m' /\ expDenote p δ = true -> aenv_correct δ m';

    analysis_monotone : forall Δ (m1 m2 : hlist A Δ) (m1' m2' : hlist A (new_var Δ)) p,
        aenv_Incl m1 m2 -> analysis (add_event_aenv m1) p = Some m1' ->
        analysis (add_event_aenv m2) p = Some m2' -> aenv_Incl m1' m2';

    analysis_monotone' : forall Δ (m1 m2 : hlist A Δ) p,
        aenv_Incl m1 m2 -> analysis (add_event_aenv m2) p = None ->
        analysis (add_event_aenv m1) p = None;
  }.

Lemma arg_env_aenv_correct A `{P : PredicateAnalysis A} : forall Δ (δ : env Δ) ts (ae : arg_env ts Δ) (m : hlist A Δ),
    aenv_correct δ m ->
    aenv_correct (arg_envDenote δ ae) (arg_env_aenv m ae).
Proof.
  dependent induction ae; simpl; intros.
  unfold aenv_correct in *.
  intros.
  inversion x.
  unfold aenv_correct in *.
  intros.
  dependent destruction x0; simpl; auto.
  destruct P.
  auto.
Qed.

Lemma add_event_aenv_correct A `{P : PredicateAnalysis A} : forall e Δ (δ : env Δ) (m : hlist A Δ),
    aenv_correct δ m ->
    aenv_correct (addEvent e δ) (add_event_aenv m).
Proof.
  intros.
  destruct e.
  simpl.
  unfold aenv_correct in *.
  intros.
  dependent destruction x.
  simpl.
  edestruct L.
  apply InIncl.
  apply Incl_top.
  dependent destruction x1.
  simpl.
  edestruct L.
  apply InIncl.
  apply Incl_top.
  dependent destruction x1.
  simpl.
  edestruct L.
  apply InIncl.
  apply Incl_top.
  dependent destruction x1.
  simpl.
  edestruct L.
  apply InIncl.
  apply Incl_top.
  simpl.
  apply H.
Qed.

Lemma aenv_Incl_arg_env_aenv A `{P : PredicateAnalysis A} : forall Δ ts (m m' : hlist A Δ) (ae : arg_env ts Δ),
    aenv_Incl m m' ->
    aenv_Incl (arg_env_aenv m ae) (arg_env_aenv m' ae).
Proof.
  dependent induction ae; simpl; intros; auto;

  unfold aenv_Incl; intros.
  inversion x.
  dependent destruction x0; simpl in *.
  apply arg_eval_monotone; auto.

  apply IHae in H.
  unfold aenv_Incl in H.
  specialize (H τ x2).
  auto.
Qed.

Class CSLAnalysis (L : Type) (A : ty -> Type) `(Lattice L) `(PredicateAnalysis A) :=
  {
    β : trace -> L;
    L__succ : L;
    β_empty : Incl (β nil) L__succ;

    C__par : L -> L -> L;
    β__par : forall t1 t2 t l1 l2,
        interleave t1 t2 t -> Incl (β t1) l1 /\ Incl (β t2) l2 -> Incl (β t) (C__par l1 l2);
    monotone_C__par : forall l1 l2 l3 l4, Incl l1 l2 /\ Incl l3 l4 -> Incl (C__par l1 l3) (C__par l2 l4);

    C__seq : L -> L -> L;
    β__seq : forall t1 t2 t l1 l2,
        appends t1 t2 t -> Incl (β t1) l1 /\ Incl (β t2) l2 -> Incl (β t) (C__seq l1 l2);
    monotone_C__seq : forall l1 l2 l3 l4, Incl l1 l2 /\ Incl l3 l4 -> Incl (C__seq l1 l3) (C__seq l2 l4);

    C__transfer : forall Δ, hlist A (new_var Δ) -> L -> L;
    β__transfer : forall Δ (δ : env Δ) (m' : hlist A (new_var Δ)) l e t,
        aenv_correct (addEvent e δ) m' ->
        Incl (β t) l -> Incl (β (cons e t)) (C__transfer m' l);
    monotone_C__transfer : forall Δ (m m' : hlist A (new_var Δ)) l l',
        aenv_Incl m m' -> Incl l l' -> Incl (C__transfer m l) (C__transfer m' l')
  }.

Inductive ind_csl_analysis L A `{CSLAnalysis L A} : forall Γ Δ, contract Γ Δ -> template_env Γ -> hlist A Δ -> L -> Prop :=
| SuccessA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ), ind_csl_analysis (success Γ Δ) D m L__succ
| FailureA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ), ind_csl_analysis (failure Γ Δ) D m bot
| AltA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    ind_csl_analysis c1 D m l1 -> ind_csl_analysis c2 D m l2
    -> ind_csl_analysis (c1 :+: c2) D m (join l1 l2)
| SeqA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    ind_csl_analysis c1 D m l1 -> ind_csl_analysis c2 D m l2 -> ind_csl_analysis (c1 ;; c2) D m (C__seq l1 l2)
| ParA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    ind_csl_analysis c1 D m l1 -> ind_csl_analysis c2 D m l2 -> ind_csl_analysis (c1 :||: c2) D m (C__par l1 l2)
| TransferAFail : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) c (p : exp (new_var Δ) Bool),
    analysis (add_event_aenv m) p = None ->
    ind_csl_analysis (transfer p c) D m bot
| TransferASucc : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (m' : hlist A (new_var Δ)) c (p : exp (new_var Δ) Bool) l,
    analysis (add_event_aenv m) p = Some m' ->
    ind_csl_analysis c D m' l ->
    ind_csl_analysis (transfer p c) D m (C__transfer m' l)
| TappA : forall Γ Δ (D : template_env Γ) ts (ae : arg_env ts Δ) (f : member ts Γ) (m : hlist A Δ) l,
    ind_csl_analysis (hget D f) D (arg_env_aenv m ae) l
    -> ind_csl_analysis (tapp ae f) D m l.

Hint Resolve arg_env_aenv_correct add_event_aenv_correct : csl.

Theorem ind_csl_analysis_sound L A `{CSLAnalysis L A} : forall Γ Δ (D : template_env Γ) (δ : env Δ) (m : hlist A Δ) (c : contract Γ Δ) r,
    aenv_correct δ m /\ ind_csl_analysis c D m r ->
    forall t, csat δ D t c -> Incl (β t) r.
Proof.
  intros.
  destruct H0, H1, H2;
  generalize dependent t.
  generalize dependent δ.
  induction H1; intros;
    match goal with
    | [ H : csat _ _ _ _ |- _ ] => dependent destruction H; eauto with csl
    end.

  apply IHind_csl_analysis1 with (t := t) in H0; eauto with csl.
  eapply Incl_trans.
  apply H0; auto.
  apply join_correct.

  apply IHind_csl_analysis2 with (t := t) in H3; eauto with csl.
  eapply Incl_trans.
  apply H3; auto.
  apply join_correct.

  contradict H2.
  rewrite Bool.not_true_iff_false.
  eapply analysis_None0; intuition.
  apply H0.
  eapply add_event_aenv_correct in H1.
  apply H1.

  eapply β__transfer; eauto with csl.
  eapply analysis_Some; eauto with csl.

  eapply IHind_csl_analysis.
  eapply analysis_Some; eauto with csl.
  auto with csl.
Qed.

Lemma env_widening_sound L A `(CSLAnalysis L A) : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ) (c : contract Γ Δ) s s',
    aenv_Incl m m' /\
    ind_csl_analysis c D m s /\
    ind_csl_analysis c D m' s' ->
    Incl s s'.
  intros.
  destruct H2 as [? [? ?]].
  generalize dependent s';
    generalize dependent m';
    induction H3; intros;
      match goal with (* Case analysis on second derivation *)
      | [ H : ind_csl_analysis _ _ _ ?s' |- Incl _ ?s' ] =>
        dependent destruction H;
          eauto
      end.

  apply Incl_refl.
  apply Incl_refl.

  apply join_monotone; eauto.

  apply monotone_C__seq; split; eauto.

  apply monotone_C__par; split; eauto.

  apply Incl_bot.
  apply Incl_bot.

  eapply analysis_monotone' in H4; eauto.
  rewrite H2 in H4.
  inversion H4.

  apply monotone_C__transfer; eauto.
  eapply analysis_monotone; eauto.
  eapply IHind_csl_analysis; auto.
  eapply analysis_monotone; eauto.
  eauto.

  eapply IHind_csl_analysis; eauto.
  apply aenv_Incl_arg_env_aenv; eauto.

Qed.

CoInductive csl_analysis L A `(CA : CSLAnalysis L A) : forall Γ Δ, contract Γ Δ -> template_env Γ -> hlist A Δ -> L -> Prop :=
| CSuccessA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ), csl_analysis CA (success Γ Δ) D m L__succ
| CFailureA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ), csl_analysis CA (failure Γ Δ) D m bot
| CAltA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    csl_analysis CA c1 D m l1 -> csl_analysis CA c2 D m l2
    -> csl_analysis CA (c1 :+: c2) D m (join l1 l2)
| CSeqA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    csl_analysis CA c1 D m l1 -> csl_analysis CA c2 D m l2 -> csl_analysis CA (c1 ;; c2) D m (C__seq l1 l2)
| CParA : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c1 c2 : contract Γ Δ) l1 l2,
    csl_analysis CA c1 D m l1 -> csl_analysis CA c2 D m l2 -> csl_analysis CA (c1 :||: c2) D m (C__par l1 l2)
| CTransferAFail : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) c (p : exp (new_var Δ) Bool),
    analysis (add_event_aenv m) p = None ->
    csl_analysis CA (transfer p c) D m bot
| CTransferASucc : forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (m' : hlist A (new_var Δ)) c (p : exp (new_var Δ) Bool) l,
    analysis (add_event_aenv m) p = Some m' ->
    csl_analysis CA c D m' l ->
    csl_analysis CA (transfer p c) D m (C__transfer m' l)
| CTappA : forall Γ Δ (D : template_env Γ) ts (ae : arg_env ts Δ) (f : member ts Γ) (m : hlist A Δ) l,
    csl_analysis CA (hget D f) D (arg_env_aenv m ae) l
    -> csl_analysis CA (tapp ae f) D m l.

Theorem csl_analysis_sound L A `(CA : CSLAnalysis L A) : forall Γ Δ (D : template_env Γ) (δ : env Δ) (m : hlist A Δ) (c : contract Γ Δ) r t,
    aenv_correct δ m /\ csl_analysis CA c D m r /\ csat δ D t c ->
    Incl (β t) r.
Proof.
  intros.
  destruct H1 as [? [? ?]].
  generalize dependent r.
  generalize dependent m.
  induction H3;
  intros.
  dependent destruction H2; apply β_empty.
  dependent destruction H2.
  eapply IHcsat.
  apply arg_env_aenv_correct; eauto.
  auto.

  dependent destruction H4.
  contradict H1.
  rewrite Bool.not_true_iff_false.
  eapply analysis_None; intuition.
  apply H4.
  apply add_event_aenv_correct; auto.



  eapply β__transfer.
  eapply analysis_Some; eauto with csl.
  eapply IHcsat; auto.
  eapply analysis_Some; intuition; eauto with csl.
  eauto with csl.

  dependent destruction H2.
  eapply Incl_trans.
  eapply IHcsat; eauto.
  apply join_correct.
  dependent destruction H2.
  eapply Incl_trans.
  eapply IHcsat; eauto.
  apply join_correct.

  dependent destruction H3; eapply β__seq; eauto.
  dependent destruction H3; eapply β__par; eauto.
Qed.

(* An attempt at defining a co-inductive predicate for saying that
   an abstract semantics tree is less than another
*)
CoInductive csl_analysis_le L A `(CA : CSLAnalysis L A) : forall Γ Δ (D : template_env Γ) (c : contract Γ Δ) (l l' : L) (m m' : hlist A Δ),
    csl_analysis CA c D m l -> csl_analysis CA c D m' l' -> Prop :=
| Le_Succ : forall Γ Δ (D : template_env Γ) (m1 m2 : hlist A Δ),
    csl_analysis_le (CSuccessA CA D m1) (CSuccessA CA D m2)
| Le_Fail : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ),
    csl_analysis_le (CFailureA CA D m) (CFailureA CA D m')
| Le_Tapp : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ)
              ts (ae : arg_env ts Δ) (f : member ts Γ) (l l' : L)
              (a : csl_analysis CA (hget D f) D (arg_env_aenv m ae) l)
              (a' : csl_analysis CA (hget D f) D (arg_env_aenv m' ae) l'),
  Incl l l' -> csl_analysis_le a a' ->
  csl_analysis_le (@CTappA L A _ _ _ _ CA Γ Δ D ts ae f m l a) (@CTappA L A _ _ _ _ CA Γ Δ D ts ae f m' l' a')
| Le_Alt : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ) (c1 c2 : contract Γ Δ)
             (l1 l2 l1' l2' : L)
             (a1 : csl_analysis CA c1 D m l1)
             (a2 : csl_analysis CA c2 D m l2)
             (a1' : csl_analysis CA c1 D m' l1')
             (a2' : csl_analysis CA c2 D m' l2'),
    Incl (join l1 l2) (join l1' l2') ->
    csl_analysis_le a1 a1' -> csl_analysis_le a2 a2' ->
    csl_analysis_le (CAltA (l1 := l1) (l2 := l2) a1 a2) (CAltA (l1 := l1') (l2 := l2') a1' a2')
| Le_Seq : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ) (c1 c2 : contract Γ Δ)
             (l1 l2 l1' l2' : L)
             (a1 : csl_analysis CA c1 D m l1)
             (a2 : csl_analysis CA c2 D m l2)
             (a1' : csl_analysis CA c1 D m' l1')
             (a2' : csl_analysis CA c2 D m' l2'),
    Incl (C__seq l1 l2) (C__seq l1' l2') ->
    csl_analysis_le a1 a1' -> csl_analysis_le a2 a2' ->
    csl_analysis_le (CSeqA (l1 := l1) (l2 := l2) a1 a2) (CSeqA (l1 := l1') (l2 := l2') a1' a2')
| Le_Trans1 : forall Γ Δ (D : template_env Γ) (m1 m2 : hlist A Δ) (m1' m2' : hlist A (new_var Δ)) (c : contract Γ (new_var Δ))
                (p : exp (new_var Δ) Bool)
             (l1 l2 : L)
             (a1 : csl_analysis CA c D m1' l1)
             (a2 : csl_analysis CA c D m2' l2)
             (pa1 : analysis (add_event_aenv m1) p = Some m1')
             (pa2 : analysis (add_event_aenv m2) p = Some m2'),
    analysis (add_event_aenv m2) p = Some m2' ->
    Incl (C__transfer m1' l1) (C__transfer m2' l2) ->
    csl_analysis_le a1 a2 ->
    csl_analysis_le (@CTransferASucc L A _ _ _ _ CA Γ Δ D m1 m1' c p l1 pa1 a1) (@CTransferASucc L A _ _ _ _ CA Γ Δ D m2 m2' c p l2 pa2 a2)
| Le_Par : forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ) (c1 c2 : contract Γ Δ)
             (l1 l2 l1' l2' : L)
             (a1 : csl_analysis CA c1 D m l1)
             (a2 : csl_analysis CA c2 D m l2)
             (a1' : csl_analysis CA c1 D m' l1')
             (a2' : csl_analysis CA c2 D m' l2'),
    Incl (C__par l1 l2) (C__par l1' l2') ->
    csl_analysis_le a1 a1' -> csl_analysis_le a2 a2' ->
    csl_analysis_le (CParA (l1 := l1) (l2 := l2) a1 a2) (CParA (l1 := l1') (l2 := l2') a1' a2').

Lemma csl_analysis_le_root_incl L A `{CA : CSLAnalysis L A} :
  forall Γ Δ (D : template_env Γ) (m m' : hlist A Δ) (c : contract Γ Δ) l l'
    (a1 : csl_analysis CA c D m l) (a2 : csl_analysis CA c D m' l'),
    csl_analysis_le a1 a2 -> Incl l l'.
Proof.
  intros.
  dependent destruction a1;
  dependent destruction a2.
  apply Incl_refl.
  apply Incl_refl.
  dependent destruction H1.
  rewrite <- x0.
  rewrite <- x1.
  apply H1.
  dependent destruction H1.
  rewrite <- x0.
  rewrite <- x1.
  apply H1.
  dependent destruction H1.
  rewrite <- x0.
  rewrite <- x1.
  apply H1.
  apply Incl_refl.
  apply Incl_bot.

  dependent destruction H1.
  rewrite <- x0.
  rewrite <- x1.
  auto.
  dependent destruction H1.
  rewrite <- x0.
  rewrite <- x1.
  auto.

  dependent destruction H1.
  auto.
Qed.

(* Simple analyzer with a step counter limit *)
Fixpoint analyzeCSL {A L Γ Δ} `{CA : CSLAnalysis L A} (n : nat) (c : contract Γ Δ) (D : template_env Γ) (m : hlist A Δ) : L :=
  match n with
  | O => bot
  | S n' =>
    match c with
    | success _ _ => L__succ
    | failure _ _ => bot
    | c1 :+: c2 => join (analyzeCSL n' c1 D m) (analyzeCSL n' c2 D m)
    | c1 ;; c2 => C__seq (analyzeCSL n' c1 D m) (analyzeCSL n' c2 D m)
    | c1 :||: c2 => C__par (analyzeCSL n' c1 D m) (analyzeCSL n' c2 D m)
    | transfer p c' =>
      match analysis (add_event_aenv m) p with
      | None => bot
      | Some m' => C__transfer m' (analyzeCSL n' c' D m')
      end
    | tapp ae f => analyzeCSL n' (hget D f) D (arg_env_aenv m ae)
    end
  end.

Definition analyzeTopLevel {A L Γ Δ} `{CA : CSLAnalysis L A} {D: template_env Γ} {c: contract Γ Δ} (n : nat)  (t : TopLevel D c ) (m : hlist A Δ) : L :=
  match t with
  | LetRec D c => analyzeCSL n c D m
  end.

Definition analyzeCSL' {A L Γ Δ} `{CA : CSLAnalysis L A} (c : contract Γ Δ) (D : template_env Γ) (m : hlist A Δ) r:=
                                      exists n, (analyzeCSL n c D m = r /\ forall n', n'>= n -> analyzeCSL n' c D m = r).

Require Import Omega.

Lemma ind_csl_analysis_analyzeCSL L A `{CA : CSLAnalysis L A} :
  forall Γ Δ (D : template_env Γ) (m : hlist A Δ) (c : contract Γ Δ) r,
    ind_csl_analysis c D m r ->
    analyzeCSL' c D m r.
Proof.
  intros; induction H1; simpl.
  - eexists (S _ ); simpl; intuition; destruct n'; simpl; inversion H1; auto.
  - eexists (S _ ); intuition. destruct n'; simpl; inversion H1; auto.
  - destruct IHind_csl_analysis1 as [n1 [IH1 IH1']].
    destruct IHind_csl_analysis2 as [n2 [IH2 IH2']].
    eexists (S (n1+n2)); subst; intuition; simpl.
    + specialize IH1' with (n1+n2). specialize IH2' with (n1+n2).
      rewrite IH1'. rewrite IH2'. reflexivity. omega. omega.
    + dependent destruction n'; [inversion H1 | ].
      specialize IH1' with n'. specialize IH2' with n'.
      rewrite <- IH1'. rewrite <- IH2'. simpl. reflexivity.
      omega. omega.
  - destruct IHind_csl_analysis1 as [n1 [IH1 IH1']].
    destruct IHind_csl_analysis2 as [n2 [IH2 IH2']].
    eexists (S (n1+n2)); subst; intuition; simpl.
    + specialize IH1' with (n1+n2). specialize IH2' with (n1+n2).
      rewrite IH1'. rewrite IH2'. reflexivity. omega. omega.
    + dependent destruction n'; [inversion H1 | ].
      specialize IH1' with n'. specialize IH2' with n'.
      rewrite <- IH1'. rewrite <- IH2'. simpl. reflexivity.
      omega. omega.
  - destruct IHind_csl_analysis1 as [n1 [IH1 IH1']].
    destruct IHind_csl_analysis2 as [n2 [IH2 IH2']].
    eexists (S (n1+n2)); subst; intuition; simpl.
    + specialize IH1' with (n1+n2). specialize IH2' with (n1+n2).
      rewrite IH1'. rewrite IH2'. reflexivity. omega. omega.
    + dependent destruction n'; [inversion H1 | ].
      specialize IH1' with n'. specialize IH2' with n'.
      rewrite <- IH1'. rewrite <- IH2'. simpl. reflexivity.
      omega. omega.
  - exists (S 0); simpl; intuition. rewrite H1; reflexivity.
    dependent destruction n'; [inversion H2 | ]; simpl; rewrite H1; reflexivity.
  - destruct IHind_csl_analysis as [n [IH IH']].
    exists (S n); intuition. simpl. rewrite H1. rewrite IH'; eauto.
    dependent destruction n'; [inversion H3 | ]; simpl; rewrite H1.
    rewrite IH'; eauto. omega.
  - destruct IHind_csl_analysis as [n [IH IH']].
    exists (S n); intuition. simpl.
    dependent destruction n'; [inversion H2 | ]; simpl.
    rewrite IH'; eauto. omega.
    Unshelve.
    exact 0.
    exact 0.
Qed.
