Require Import List Program CSL.HList CSL.Definitions.

Set Implicit Arguments.

Class Lattice (L : Type) :=
  {
    top : L; bot : L;
    Incl : L -> L -> Prop;
    Incl_bot : forall (x : L), Incl bot x;
    Incl_top : forall (x : L), Incl x top;
    Incl_refl : forall (x : L), Incl x x;
    Incl_trans : forall (x y z : L), Incl x y -> Incl y z -> Incl x z;
    join : L -> L -> L;
    join_correct : forall (x y : L),
        Incl x (join x y) /\ Incl y (join x y);
    join_monotone : forall (x x' y y' : L),
        Incl x x' -> Incl y y' -> Incl (join x y) (join x' y')
  }.

Class SetLattice (A : ty -> Type) `{L : forall τ : ty, Lattice (A τ)} :=
  {
    In {τ} : tyDenote τ -> A τ -> Prop;
    lift {τ} : tyDenote τ -> A τ;
    InIncl : forall τ (v : tyDenote τ) l, In v l <-> Incl (lift v) l
  }.

(* Fixpoint join_all (A : ty -> Type) Δ (join : forall  : hlist A Δ -> hlist A Δ -> hlist A Δ :=
  match Δ return hlist A Δ -> hlist A Δ -> hlist A Δ with
  | nil => fun _ _ => HNil
  | τ :: Δ' => fun m1 m2 => HCons (join (@hhead _ A _ _ m1) (hhead m2)) (join_all (htail m1) (htail m2))
  end. *)

(*
Instance hlist_lattice {A : ty -> Type} {Δ} `(LB : forall τ, Lattice (A τ)) : Lattice (hlist A Δ) :=
  {
    top := (fix F Δ : hlist A Δ :=
              match Δ with
              | nil => HNil
              | τ :: Δ' => HCons top (F Δ')
              end) Δ;
    bot := (fix F Δ : hlist A Δ :=
              match Δ with
              | nil => HNil
              | τ :: Δ' => HCons bot (F Δ')
              end) Δ;
    Incl m m' := forall τ (x : member τ Δ), Incl (hget m x) (hget m' x);
  }.
*)

Program Instance fun_lattice {A B} `(LB : Lattice B) : Lattice (A -> B) :=
  {
    top := fun _ => top;
    bot := fun _ => bot;
    Incl x y := forall a, Incl (x a) (y a);
    join x y := fun a => join (x a) (y a)
  }.
Next Obligation.
  destruct LB.
  intros; simpl in *; eauto.
Defined.
Next Obligation.
  destruct LB.
  intros; simpl in *; eauto.
Defined.
Next Obligation.
  destruct LB.
  intros; simpl in *; eauto.
Defined.
Next Obligation.
  destruct LB.
  intros; simpl in *; eauto.
Defined.
Next Obligation.
  destruct LB.
  split; intros; edestruct join_correct0; eauto.
Defined.
Next Obligation.
  destruct LB.
  intros; simpl in *; eauto.
Defined.
