Require Import List ListSet Program.

Require Import
        CSL.HList
        CSL.Syntax
        CSL.Exp
        CSL.Definitions
        CSL.Satisfaction
        CSL.Analysis.Lattice
        CSL.Analysis.AbstractValues
        CSL.Analysis.Participation
        CSL.Analysis.Generic
        CSL.Analysis.IdPredicateAnalysis
        CSL.Analysis.Algorithms.
Set Implicit Arguments.

Hint Unfold aap_set_Incl : csl.
Hint Resolve set_add_elim set_add_intro aap_set_union_intro : csl.

(* We get all the important proofs from CSL.Analysis.Participation *)

Program Instance aap_set_lattice : Lattice (set abstract_agent_pair) :=
  { top := singleton dec_eq_aap (AnyValue, AnyValue);
    bot := empty_set _;
    Incl := aap_set_Incl;
    join := set_union dec_eq_aap
  }.
Next Obligation.
  unfold aap_set_Incl.
  destruct a.
  destruct a, a0; simpl; intuition.
Defined.
Next Obligation.
  unfold aap_set_Incl.
  intros.
  destruct a.
  destruct a, a0; simpl; intuition.
Defined.
Next Obligation.
  eauto with csl.
Defined.
Next Obligation.
  eauto with csl.
Defined.
Next Obligation.
  intuition; eauto with csl.
Defined.
Next Obligation.
  unfold aap_set_Incl.
  intros.
  destruct a.
  apply aap_set_union_elim in H1.
  apply aap_set_union_intro. intuition.
Defined.

Hint Resolve aap_set_β_participation aap_set_In_weak β_participation_app_union β_participation_interleave_union : csl.

Lemma β_append : forall t1 t2 t l1 l2,
    appends t1 t2 t -> aap_set_Incl (β_participation t1) l1 /\ aap_set_Incl (β_participation t2) l2 -> aap_set_Incl (β_participation t) (set_union dec_eq_aap l1 l2).
Proof.
  unfold aap_set_Incl;
  intros.
  destruct H0; auto.
  eapply β_participation_app_union in H; eauto with csl.
  apply aap_set_β_participation in H1; eauto with csl.
  apply set_union_elim in H.
  destruct H; auto with csl.
Qed.

Lemma β_append_Incl_R : forall t1 t2 t,
    appends t1 t2 t -> aap_set_Incl (set_union dec_eq_aap (β_participation t1) (β_participation t2)) (β_participation t) .
Proof.
  unfold aap_set_Incl.
  intros.
  induction H; simpl in *.
  - apply aap_set_union_elim in H0; intuition;
    destruct a as [a0 a1]; destruct a0; destruct a1; simpl in *; intuition.
  - autounfold with csl in *.
    apply aap_set_union_elim in H0; intuition; [apply aap_set_add_elim in H1; intuition | ];
    apply aap_set_In_weak; apply set_add_intro; right; apply aap_set_β_participation; apply IHappends;
      apply aap_set_union_intro; intuition.
Qed.

Lemma β_append_Incl_L : forall t1 t2 t,
    appends t1 t2 t -> aap_set_Incl (β_participation t) (set_union dec_eq_aap (β_participation t1) (β_participation t2)).
Proof.
  unfold aap_set_Incl.
  intros.
  induction H; simpl in *.
  - apply aap_set_union_intro. intuition.
  - autounfold with csl in *.
    apply aap_set_union_intro.
    apply aap_set_add_elim in H0; intuition.
    apply aap_set_union_elim in H0; intuition.
Qed.

Lemma β_interleave : forall t1 t2 t l1 l2,
    interleave t1 t2 t -> aap_set_Incl (β_participation t1) l1 /\ aap_set_Incl (β_participation t2) l2 -> aap_set_Incl (β_participation t) (set_union dec_eq_aap l1 l2).
Proof.
  unfold aap_set_Incl;
  intros.
  destruct H0; auto.
  eapply β_participation_interleave_union in H; eauto with csl.
  apply aap_set_β_participation in H1; eauto.
  apply set_union_elim in H.
  destruct H; auto with csl.
Qed.

Lemma β_interleave_Incl_R : forall t1 t2 t,
    interleave t1 t2 t -> aap_set_Incl (set_union dec_eq_aap (β_participation t1) (β_participation t2)) (β_participation t) .
Proof.
  unfold aap_set_Incl.
  intros.
  induction H; simpl in *; auto.
  - apply aap_set_union_elim in H0; intuition;
    destruct a as [a0 a1]; destruct a0; destruct a1; simpl in *; intuition.
  - autounfold with csl in *.
    apply aap_set_union_elim in H0; intuition; [apply aap_set_add_elim in H1; intuition | ];
    apply aap_set_In_weak; apply set_add_intro; right; apply aap_set_β_participation; apply IHinterleave;
      apply aap_set_union_intro; intuition.
  - autounfold with csl in *.
    apply aap_set_union_elim in H0; intuition.
    + apply aap_set_In_weak; apply set_add_intro; right; apply aap_set_β_participation; apply IHinterleave;
      apply aap_set_union_intro; intuition.
    + apply aap_set_add_elim in H1; intuition;
      apply aap_set_In_weak; apply set_add_intro; right; apply aap_set_β_participation; apply IHinterleave;
      apply aap_set_union_intro; intuition.
Qed.

Lemma β_interleave_Incl_L : forall t1 t2 t,
    interleave t1 t2 t -> aap_set_Incl  (β_participation t) (set_union dec_eq_aap (β_participation t1) (β_participation t2)) .
Proof.
  unfold aap_set_Incl.
  intros.
  induction H; simpl in *; auto.
  - apply aap_set_union_intro; intuition.
  - autounfold with csl in *.
    apply aap_set_union_intro; apply aap_set_add_elim in H0; intuition. apply aap_set_union_elim in H0; intuition.
  - autounfold with csl in *.
    apply aap_set_union_intro; apply aap_set_add_elim in H0; intuition. apply aap_set_union_elim in H0; intuition.
Qed.

Definition transfer_analysis Δ (m : value_map (new_var Δ)) l :=
  set_union dec_eq_aap (cross (hget m HFirst) (hget m (HNext HFirst))) l.

Lemma transfer_correct `{@PredicateAnalysis _ _ abstract_set_setlattice} : forall Δ (δ : env Δ) (m' : value_map (new_var Δ)) l e t,
        aenv_correct (addEvent e δ) m' ->
        aap_set_Incl (β_participation t) l ->
        aap_set_Incl (β_participation (e :: t)) (transfer_analysis m' l).
Proof.
  unfold aap_set_Incl, transfer_analysis.
  intros.
  apply aap_set_β_participation in H2.
  destruct e.
  simpl in *.
  unfold β_participation_folder in *.
  apply set_add_elim in H2;
  destruct H2; eauto with csl.
  subst.
  apply aap_set_union_intro.
  left.
  dependent destruction m'.
  dependent destruction m'.
  remember (H0 Agent HFirst) as Hsender.
  remember (H0 Agent (HNext HFirst)) as Hreceiver.
  apply abstract_set_In_cross; auto with csl.
Qed.

Lemma participation_join_monotone `{PredicateAnalysis abstract_set} : forall l1 l2 l3 l4,
    aap_set_Incl l1 l2 /\ aap_set_Incl l3 l4 -> aap_set_Incl (join l1 l3) (join l2 l4).
Proof.
  intros.
  intuition.
  unfold aap_set_Incl in *.
  intros.
  apply aap_set_union_elim in H0.
  destruct H0;
  apply aap_set_union_intro; auto.
Qed.

Lemma aap_set_In_cross `{@PredicateAnalysis _ _ abstract_set_setlattice} : forall Δ (m m' : value_map (new_var Δ)) a,
    aenv_Incl m m' ->
    aap_set_In a (cross (hget m HFirst) (hget m (HNext HFirst))) ->
    aap_set_In a (cross (hget m' HFirst) (hget m' (HNext HFirst))).
Proof.
  intros.

  dependent destruction m; dependent destruction m.
  dependent destruction m'; dependent destruction m'.
  remember (H0 Agent HFirst).
  remember (H0 Agent (HNext HFirst)).
  simpl in *.
  eapply aap_set_in_cross_help; eauto.
Qed.

Lemma transfer_analysis_monotone `{@PredicateAnalysis _ _ abstract_set_setlattice} : forall (Δ : list ty) (m m' : value_map (new_var Δ)) l l',
    aenv_Incl m m' -> Incl l l' -> Incl (transfer_analysis m l) (transfer_analysis m' l').
Proof.
  intros.
  simpl in *.
  unfold aap_set_Incl in *.
  intros.
  unfold transfer_analysis in *.
  apply aap_set_union_elim in H2.
  destruct H2;
  apply aap_set_union_intro.
  left.
  eapply aap_set_In_cross; try apply H0; auto.
  right; eauto.
Qed.

Program Instance participation_analysis `{P : @PredicateAnalysis _ _ abstract_set_setlattice } : CSLAnalysis aap_set_lattice P :=
  {
    β := β_participation;
    L__succ := bot;
    C__par := join; C__seq := join;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := participation_join_monotone;
    monotone_C__seq := participation_join_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.
Next Obligation.
  auto with csl.
Defined.

Program Instance participation_id_predicate_analysis:
  CSLAnalysis aap_set_lattice id_predicate_analysis_abstract_set :=
  {
    β := β_participation;
    L__succ := bot;
    C__par := join; C__seq := join;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := participation_join_monotone;
    monotone_C__seq := participation_join_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.
Next Obligation.
 auto with csl.
Defined.

Program Instance participation_possible_values :
  CSLAnalysis aap_set_lattice possible_values_predicate_analysis :=
  {
    β := β_participation;
    L__succ := bot;
    C__par := join; C__seq := join;
    C__transfer := transfer_analysis;
    β__transfer := transfer_correct;
    β__par := β_interleave;
    β__seq := β_append;
    monotone_C__par := participation_join_monotone;
    monotone_C__seq := participation_join_monotone;
    monotone_C__transfer := transfer_analysis_monotone
  }.
Next Obligation.
  auto with csl.
Defined.

Require Import String.
Open Scope string_scope.

(* This is an autogenerated example created by a simple program that parses (the de-sugaring of)
   the CSL code given below, and produces the corresponding Coq AST, with the deadline given as
   an arbitrary integer (i.e. not number of seconds from 1/1/1970)

letrec
  sale[trusted, seller, buyer, goods, payment, deadline] =
    Transfer(buyer, trusted, payment, _).
    (Transfer(seller, buyer, goods, T | T < deadline).
    Transfer(trusted, seller, payment, T’ | True).Success
    + Transfer(trusted, buyer, payment, T | T > deadline).Success)
  in sale("3rd", "shop", "alice", 1 bike, 1000 EUR, 2019-09-01)
*)
Example exampleSell :=
 LetRec (Δ:=[]) (Γ := (cons ((cons Agent (cons Agent (cons Agent (cons Resource (cons Resource (cons Timestamp []))))))) []))
	(HCons (transfer (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext HFirst))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))) (alternate (transfer (Binop And (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Binop (Le TimestampCmp ) (Var (HNext (HNext (HNext HFirst)))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))))) (transfer (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))))))))) (@success _ _))) (transfer (Binop And (Binop And (Binop And (Binop (Eq AgentEq ) (Var HFirst) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))) (Binop (Eq AgentEq ) (Var (HNext HFirst)) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))) (Binop (Eq ResourceEq ) (Var (HNext (HNext HFirst))) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst))))))))))))))) (Binop (Le TimestampCmp ) (Var (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext (HNext HFirst)))))))))))))) (Var (HNext (HNext (HNext HFirst)))))) (@success _ _)))) HNil)
	(tapp (HCons (@Lit _ Agent "3rd") (HCons (@Lit _ Agent "shop") (HCons (@Lit _ Agent "alice") (HCons (@Lit _ Resource [("bike",1)]) (HCons (@Lit _ Resource [("eur",1000)]) (HCons (@Lit _ Timestamp 2019) HNil)))))) HFirst).

Eval compute in (analyzeTopLevel (CA:=participation_possible_values) 500 exampleSell HNil).
