Require Import List ListSet Program.
Require Import CSL.Exp
        CSL.Definitions.

Definition dec_eq_agent : forall (x y : tyDenote Agent), {x = y} + {x <> y}.
  repeat (decide equality; subst).
Defined.

Fixpoint is_singleton_set (s : set (tyDenote Agent)) :=
  match s with
  | [] => false
  | (a :: []) => true
  | a :: s' => andb (set_mem dec_eq_agent a s') (is_singleton_set s') (* Works with multiple copies of the same element *)
  end.

Lemma is_singleton_set_unique_In:
  forall s,
    is_singleton_set s = true ->
    exists! a, set_In a s.
Proof.
  induction s; simpl in *; intuition.
  destruct s; intuition.
  + exists a. unfold unique; intuition.
  + apply andb_prop in H; intuition. destruct H. exists x; unfold unique in *; intuition.
    subst. apply H3. eapply set_mem_correct1; eauto.
Qed.

Lemma singleton_set_equiv_single_elem_mem:
  forall (s : set (tyDenote Agent)) a,
    is_singleton_set s = true ->
    set_In a s ->
    set_mem dec_eq_agent a s = set_mem dec_eq_agent a [a].
Proof.
  intuition. simpl. destruct (dec_eq_agent a a); [ | contradiction].
  apply is_singleton_set_unique_In in H. destruct H.
  apply set_mem_correct2; assumption.
Qed.

Lemma is_singleton_set_In_eq:
  forall s s0,
    is_singleton_set (s :: s0) = true -> forall v, set_In v (s::s0) -> v = s.
Proof.
  intros. apply is_singleton_set_unique_In in H. destruct H. unfold unique in H.
  intuition. assert (x = v) by intuition. assert (x = s). apply H2. simpl; tauto. rewrite <- H. rewrite <- H3. reflexivity.
Qed.

Lemma not_singleton_cons:
  forall v s,
    is_singleton_set (v :: v :: s) = false -> is_singleton_set (v :: s) = false.
Proof.
  simpl in *; intuition.
  destruct (dec_eq_agent v v); intuition.
Qed.

Lemma not_singleton:
  forall s v, is_singleton_set (v :: s) = false -> exists v', set_In v' (v :: s) /\ v <> v'.
Proof.
  induction s; intuition.
  destruct (dec_eq_agent v a).
  + subst. apply not_singleton_cons in H. destruct IHs with a; intuition.
    exists x; simpl; intuition.
  + exists a; simpl; intuition.
Qed.
