(** printing τ $\tau$ #τ# *)
(** printing τ' $\tau'$ #τ'# *)
(** printing τ1 $\tau_1$ #τ2# *)
(** printing τ2 $\tau_2$ #τ2# *)
(** printing Δ $\Delta$ #Δ# *)
(** printing Δ' $\Delta'$ #Δ'# *)
(** printing δ $\delta$ #δ# *)
(** printing δ' $\delta'$ #δ'# *)
(** printing Γ $\Gamma$ #Γ# *)
Require Import String List Bool.
Require Import CSL.HList.

Import ListNotations.

Set Implicit Arguments.

Inductive ty : Set :=  Agent | Resource | Timestamp | Bool | Scale.

Scheme Equality for ty.

Definition ty_env := list ty.

Definition tyDenote (t : ty) : Set :=
  match t with 
  | Agent => string
  | Resource => list (string * nat)
  | Timestamp => nat
  | Bool => bool
  | Scale => nat
  end.

Definition tyDenote_dec_eq τ : forall x y : tyDenote τ, {x = y} + {x <> y}.
  destruct τ;
  intros;
  repeat decide equality.
Defined.

Definition testResource (f : nat -> nat -> bool) (xs: tyDenote Resource) (ys: tyDenote Resource) : bool :=
  List.forallb (
      fun z => match z with
                 ( (nm1, v1), (nm2, v2)) => String.eqb nm1 nm2 && f v1 v2
               end ) (List.combine xs ys) && Nat.eqb (List.length xs) (List.length ys).

Definition resourceEqb := testResource Nat.eqb.
Definition resourceLeb := testResource Nat.leb.

Lemma resourceEqb_refl:
  forall r : tyDenote Resource, resourceEqb r r = true.
Proof.
  simpl; induction r; unfold resourceEqb, testResource; simpl; [reflexivity | ].
  destruct a; simpl.
  apply andb_true_intro; split; [ apply andb_true_intro; split; [apply andb_true_intro; split | ] | ];
   [apply String.eqb_refl | apply PeanoNat.Nat.eqb_refl | apply andb_prop in IHr; intuition |  apply PeanoNat.Nat.eqb_refl].
Qed.

Lemma resourceEqb_eq:
  forall r1 r2 : tyDenote Resource, resourceEqb r1 r2 = true <-> r1 = r2.
Proof.
  simpl; unfold resourceEqb, testResource; induction r1; destruct r2; simpl; split; intros;
    try reflexivity; try discriminate.
  - destruct a; destruct p; repeat (apply andb_prop in H; destruct H);
      rewrite String.eqb_eq in H; rewrite PeanoNat.Nat.eqb_eq in H2; subst; destruct IHr1 with (r2:=r2);
        rewrite H; [ reflexivity |  apply andb_true_intro; split; auto].
  - repeat (apply andb_true_intro; split).
    + destruct a; destruct p; inversion H; subst.
      apply andb_true_intro; split; [rewrite String.eqb_refl | rewrite PeanoNat.Nat.eqb_eq]; reflexivity.
    + inversion H; subst. destruct IHr1 with (r3:=r2).
      apply andb_prop in H1; intuition.
    + inversion H; subst; rewrite PeanoNat.Nat.eqb_eq; reflexivity.
Qed.

Lemma resourceEqb_neq:
  forall r1 r2 : tyDenote Resource, resourceEqb r1 r2 = false <-> r1 <> r2.
Proof.
  simpl; unfold resourceEqb, testResource; induction r1; destruct r2; simpl in *; split; intros;
    intuition; try discriminate.
  - destruct a; destruct p. inversion H0; subst.   
    repeat (apply andb_false_elim in H; destruct H as [H | H]);
      [rewrite String.eqb_refl in H | rewrite PeanoNat.Nat.eqb_refl in H | |  rewrite PeanoNat.Nat.eqb_refl in H]; try discriminate.
    specialize (IHr1 r2). destruct IHr1. apply H1; [ | reflexivity];
    rewrite PeanoNat.Nat.eqb_refl; rewrite andb_true_r; assumption.
  - destruct a; destruct p. 
    assert (s <> s0 \/ n <> n0 \/ r1 <> r2). 
    +  destruct (String.string_dec s s0); [subst; right | left; assumption]. 
       destruct (PeanoNat.Nat.eq_dec n n0); [subst; right | left; assumption].
       intro; elim H; subst; reflexivity.
    + intuition.
      * rewrite <- andb_assoc; rewrite <- andb_assoc; apply andb_false_iff; left; apply String.eqb_neq; assumption.
      * rewrite <- andb_assoc; apply andb_false_iff; left; apply andb_false_iff; right. apply PeanoNat.Nat.eqb_neq; assumption.
      * rewrite <- andb_assoc; apply andb_false_iff. right. 
        apply IHr1 in H0. assumption.
Qed.

Lemma tyDenote_dec_eq_refl:
  forall A τ (e : tyDenote τ) (e1 e2 : A), (if (tyDenote_dec_eq τ e e) then e1 else e2) = e1.
Proof.
  intros; destruct τ;
  destruct (tyDenote_dec_eq _ e e); try contradiction; auto.
Qed.
  
Inductive event : Set :=
| Event : tyDenote Agent -> tyDenote Agent -> tyDenote Resource -> tyDenote Timestamp -> event.

Definition trace := list event.

Definition trace_dec_eq : forall (e1 e2 : trace), {e1 = e2} + {e1 <> e2}.
  repeat decide equality.
Defined.

Definition env Δ := hlist tyDenote Δ.

Definition new_var env := Agent :: Agent :: Resource :: Timestamp :: env.

Definition addEvent (e : event) (Δ : list ty) (δ : env Δ) : env (new_var Δ) :=
  match e in event return env (new_var Δ) with
  | Event a1 a2 r t =>
    @HCons ty tyDenote Agent _ a1
           (@HCons ty tyDenote Agent _ a2
                   (@HCons ty tyDenote Resource _ r
                           (@HCons ty tyDenote Timestamp Δ t δ)))
  end.
