Require Import List Program.
Import ListNotations.

Set Implicit Arguments.

Section hlist.
  Variable A : Type.
  Variable B : A -> Type.

  Inductive hlist : list A -> Type :=
  | HNil : hlist nil
  | HCons : forall (x : A) (ls : list A), B x -> hlist ls -> hlist (x :: ls).

  Variable elm : A.

  Inductive member : list A -> Type :=
  | HFirst : forall ls, member (elm :: ls)
  | HNext : forall x ls, member ls -> member (x :: ls).

  Fixpoint hget ls (mls : hlist ls) : member ls -> B elm :=
    match mls with
    | HNil
      => fun mem =>
           match mem in member ls'
                 return (match ls' with
                         | nil => B elm
                         | _ :: _ => unit
                         end) with
           | HFirst _ => tt
           | HNext _ _ => tt
           end
    | HCons x mls'
      => fun mem =>
           match mem in member ls'
                 return (match ls' with
                         | nil => Empty_set
                         | x' :: ls'' =>
                           B x' -> (member ls'' -> B elm)
                           -> B elm
                         end) with
           | HFirst _ => fun x _ => x
           | HNext  _ mem' => fun _ get_mls' => get_mls' mem'
           end x (hget mls')
    end.

  Program Fixpoint hput ls (e : B elm) (mls : hlist ls) (m : member ls) : hlist ls :=
    match mls with
    | HNil => HNil
    | HCons x mls' => match m with
                     | HFirst _ => HCons e mls'
                     | HNext _ mem' => HCons x (hput e mls' mem')
                     end
    end.

  Lemma hput_hget_id : forall ls (m : member ls) (mls : hlist ls),
      hput (hget mls m) mls m = mls.
  Proof.
    intros.
    dependent induction m; dependent destruction mls; auto.
    simpl.
    rewrite IHm; auto.
  Qed.

  Lemma hget_hput : forall ls (e : B elm) (mls : hlist ls) (m : member ls),
      hget (hput e mls m) m = e.
  Proof.
    intros.
    dependent induction m; dependent destruction mls; auto.
    simpl; rewrite IHm; auto.
  Qed.

  Definition hhead e ls (mls : hlist (e :: ls)) : B e :=
    match mls with
    | HCons x _ => x
    end.

  Definition htail e ls (mls : hlist (e :: ls)) : hlist ls :=
    match mls with
    | HCons x mls' =>  mls'
    end.

  Lemma hhead_htail : forall e ls (mls : hlist (e :: ls)),
      HCons (hhead mls) (htail mls) = mls.
    Proof.
      intros.
      dependent destruction mls.
      auto.
    Qed.

  Section hget_dec.
    Hypothesis elm_dec : forall (v1 v2 : B elm), {v1 = v2} + {v1 <> v2}.

    Lemma hget_dec : forall ls (mls : hlist ls) (m : member ls) (v : B elm), {hget mls m = v} + {hget mls m <> v}.
    Proof.
      intros.
      dependent induction m;
        dependent destruction  mls.
      simpl.
      apply elm_dec.
      simpl.
      apply IHm.
    Qed.

  End hget_dec.
End hlist.

Arguments HNil {A B}.
Arguments HCons [A B x ls].

Arguments HFirst {A elm ls}.
Arguments HNext [A elm x ls].

Section happend.

  Variable A : Type.
  Variable B : A -> Type.

  Fixpoint happend ls1 ls2 (mls1 : hlist B ls1) (mls2 : hlist B ls2) : hlist B (ls1 ++ ls2) :=
    match mls1 with
    | HNil => mls2
    | HCons a mls1' => HCons a (happend mls1' mls2)
    end.


End happend.

Notation "[| e ; .. ; f |]" := (HCons e .. (HCons f HNil) ..).


Lemma member_dec_eq:
  forall A τ (Δ : list A) (v0 v1: member τ Δ), {v0 = v1} + {v0 <> v1}.
Proof.
  intros. induction v0; dependent destruction v1. left; reflexivity. right; discriminate. right; discriminate. destruct (IHv0 v1); [subst; left | right]; auto.
  intro Hn; inversion Hn. simpl_existTs; subst; contradiction.
Qed.
