Require Import List.

Import ListNotations.

Require Import CSL.Syntax.
Require Import CSL.Definitions.
Require Import CSL.Exp.
Require Import CSL.HList.

Set Implicit Arguments.

Inductive interleave (T : Set) : list T -> list T -> list T -> Prop :=
| IntLeftNil  : forall t, interleave nil t t
| IntRightNil : forall t, interleave t nil t
| IntLeftCons : forall t1 t2 t3 e, interleave t1 t2 t3 -> interleave (e :: t1) t2 (e :: t3)
| IntRightCons : forall t1 t2 t3 e, interleave t1 t2 t3 -> interleave t1 (e :: t2) (e :: t3).

Inductive appends (T : Set) : list T -> list T -> list T -> Prop :=
| AppNil  : forall t, appends nil t t
| AppCons : forall t l r h, appends t l r -> appends (h :: t) l (h :: r).

Inductive csat : forall Γ Δ, env Δ -> template_env Γ -> trace -> contract Γ Δ -> Prop :=
| NilSuccess : forall Γ Δ D δ, csat D δ nil (success Γ Δ) 
| AppF : forall ts Γ Δ δ (D : template_env Γ) (ae : arg_env ts Δ) f t,
    csat (arg_envDenote δ ae) D t (hget D f) -> csat δ D t (tapp ae f)
| Transfer : forall Γ Δ (D : template_env Γ) δ e (p : exp (new_var Δ) Bool) t c,
    csat (addEvent e δ) D t c -> (* bind the first event in the rest of the contract *)
    expDenote p (addEvent e δ) = true -> (* predicate is true *)
    csat δ D (e :: t) (transfer p c)
| LeftAlt : forall Γ Δ (δ : env Δ) (D : template_env Γ) t c1 c2,
    csat δ D t c1 -> csat δ D t (alternate c1 c2)
| RightAlt : forall Γ Δ (δ : env Δ) (D : template_env Γ) t c1 c2,
    csat δ D t c2 -> csat δ D t (alternate c1 c2)
| SeqSat : forall Γ Δ (δ : env Δ) (D : template_env Γ) t1 t2 t3 c1 c2,
    csat δ D t1 c1 -> csat δ D t2 c2 -> appends t1 t2 t3 -> csat δ D t3 (sequence c1 c2)
| ConcSat : forall Γ Δ (δ : env Δ) (D : template_env Γ) t t1 t2 c1 c2,
    csat δ D t1 c1 -> csat δ D t2 c2 -> interleave t1 t2 t -> csat δ D t (concurrent c1 c2).

(*
Notation "Env , D |- , Trace : Contract" := (csat Env D Trace Contract) (at level 100, no associativity).
*)
